#include "macro-lib.cc"
#include <iomanip>
#include <fstream>

Point dividerOffset = { 10, 49 };
Point boxOffset = { 10, 132 - 266 }; // Offset from bottom-left

Point beginOffset = { 95, 102 };
Point addWoodOffset = { 26, 108 };
Point addWaterOffset = { 76, 108 };
Point oxyLowOffset = { 120, 108 };
Point oxyMedOffset = { 145, 108 };
Point oxyHighOffset = { 170, 108 };

Point heatBarOffset = { 64, 13 };
Point oxyBarOffset = { 64, 25 };
Point woodBarOffset = { 64, 37 };
Point waterBarOffset = { 64, 49 };
Point dangerBarOffset = { 64, 61 };
Point progressBarOffset = { 64, 73 };

const int maxBarLength = 116;


struct HearthMenu {  
  Point topLeft;
  Point boxTopLeft;
  double lastWood;
  int oxyLevel;
  bool hasStarted;
  bool done;
  int addedWood;

  HearthMenu(Rectangle * rect) {
    topLeft = rect->topLeft;
    boxTopLeft = topLeft + boxOffset;
    boxTopLeft.y += rect->size.y;
    init();
    cout << "Menu at " << topLeft << " with box at " << boxTopLeft << endl;
  }

  void init() {
    oxyLevel = 1;
    lastWood = -1;
    hasStarted = false;
    done = false;
    addedWood = 0;
  }

  void press(Point offset) {
    Point p = boxTopLeft + offset;
    clickWait(p);
  }

  double getBarLength(Point offset) {
    Point p = boxTopLeft + offset;
    return (double)barLength(tale.w, p) / maxBarLength;
  }

  double getProgressBarLength() {
    Point p = boxTopLeft + progressBarOffset;
    Color c = getColor(tale.w, p);
    if (c.g > c.b && c.g > c.r) {
      return 1.0;
    }
    return (double)barLength(tale.w, p) / (maxBarLength + 1);
  }

  void dumpLengths() {
    ios::fmtflags f = cout.flags();
    cout << setprecision(3) << fixed;
    cout << getBarLength(heatBarOffset) << "\t";
    cout << getBarLength(oxyBarOffset) << "\t";
    cout << getBarLength(woodBarOffset) << "\t";
    cout << getBarLength(waterBarOffset) << "\t";
    cout << getBarLength(dangerBarOffset) << "\t";
    cout << getBarLength(progressBarOffset) << endl;
    cout.flags(f);
  }
  bool hasChanged() {
    if (done) return false;
    double curWood = getBarLength(woodBarOffset);
    bool hasChanged = (fabs(curWood - lastWood) > 0.005);
    lastWood = curWood;
    return hasChanged;
  }

  void increaseOxy() {
    if (oxyLevel < 2) {
      setOxy(oxyLevel + 1);
    }
  }

  void setOxy(int level) {
    if (level == oxyLevel) {
      return;
    }
    oxyLevel = level;
    if (level == 0) {
      press(oxyLowOffset);
    } else if (level == 2) {
      press(oxyHighOffset);
    } else {
      press(oxyMedOffset);
    }
  }

  void addWater() {
    press(addWaterOffset);
  }

  void addWood() {
    addedWood++;
    press(addWoodOffset);
  }

  double getPotential() {
    // A measure of how high the oven danger could go at current readings
    double wood = getBarLength(woodBarOffset);
    double water = getBarLength(waterBarOffset);
    double danger = getBarLength(dangerBarOffset);
    return (wood - water * 2) * (1 + wood) + danger;
  }
  
  void adjust() {
    if (done) return;
    double heat = getBarLength(heatBarOffset);

    if (heat > 0) {
      // Check for already running oven
      hasStarted = true;
    }
    if (!hasStarted) {
      // If not yet started and isn't already running, start it
      press(beginOffset);
      waitForRefresh();
      addWood();
      addWood();
      return;
    }
    // Check to see whether it just finished
    if (hasStarted && heat == 0) {
      done = true;
      return;
    }

    double oxy = getBarLength(oxyBarOffset);
    double wood = getBarLength(woodBarOffset);
    double danger = getBarLength(dangerBarOffset);
    double progress = getProgressBarLength();
    double potential = getPotential();

    if (danger > 0.95 && potential > 1.1) {
      // Prevent oven from blowing up.  This should normally not happen.
      addWater();
      potential = getPotential();
    }

    if (progress >= 1) {
      // Oven is finished; try to cool it quickly
      if (danger < 0.8 || potential < 0.9 || oxy >= danger) {
	setOxy(2);
      } else {
	setOxy(1);
      }
      if (danger > 0.8 && potential > 0.95) {
	addWater();
      }
      return;
    }

    double coast = progress + ((heat > 0.5) ? 2 * pow((heat-0.5), 2) : 0);

    // Add wood if needed
    if (coast < 1 && potential < 0.96) {
      addWood();
      wood = getBarLength(woodBarOffset);
    }

    // Keep oxygen low but avoid snuffing.
    if (oxy < wood * 0.2) {
      setOxy(2);
    } else if (oxy < wood * 0.6) {
      setOxy(1);
    } else {
      setOxy(0);
    }
  }
};

const int maxMenus = 4;

int main(int argc, char *argv[]) {
  initMacro();

  RectList * rects = getMenus();
  int numMenus = rects->size();

  if (numMenus < 1) {
    printf("Couldn't find usable menus!  Aborting\n");
    endMacro();
    exit(1);
  }
  
  HearthMenu * hearths[numMenus];
  for (int i = 0; i < numMenus; i++) {
    hearths[i] = new HearthMenu((*rects)[i]);
  }
  
  int numActiveMenus = numMenus;
  while (numActiveMenus > 0) {
    numActiveMenus = 0;
    for (int i = 0; i < numMenus; i++) {
      if (!(hearths[i]->done)) {
	++numActiveMenus;
	if (hearths[i]->hasChanged()) {
	  hearths[i]->adjust();
	}
      }
    }
    waitForRefresh();
  }

  endMacro();

  cout << "Time elapsed: " << getMillis() << " ms" << endl;
  for (int i = 0; i < numMenus; i++) {
    cout << "Hearth " << i << " added " << hearths[i]->addedWood << " wood" << endl;
  }
}
