#include "macro-lib.cc"
#include "reactory-model.cc"
#include <iomanip>

Window copyWindow() {
  Window w = XCreateWindow(dpy, tale.w,
			   0, 0,
			   tale.width, tale.height,
			   0, 0, InputOutput,
			   0, 0, 0);
  XMapWindow(dpy, w);
  return w;
}

const float maxAlpha = 0.5;
const float radius = 128.4;
const int maxCircles = 100;

struct Reactory {
  Point topLeft;
  Point size;
  int width;
  float * oldMap;
  int numCircles;
  Point circles[maxCircles];

  Reactory() {
    oldMap = 0;
    setBounds();
  }

  void setBounds() {
    if (oldMap) {
      delete oldMap;
      oldMap = 0;
    }

    // Find extent of reactory area

    // Search upward from centre, looking for first large contiguous
    // gray-black area
    XImage * img = getImage((Point) {tale.width/2, 0}, 1, tale.height/2);
    int minHeight = 100;
    int y = tale.height / 2 - 1;
    int maxY = y;
    bool prevGray = false;
    while (y >= 0 && maxY - y < minHeight) {
      if (Color::fromImage(img, 0, y).saturation() < 0.01) {
	if (!prevGray) {
	  maxY = y;
	}
	prevGray = true;
      } else {
	prevGray = false;
      }
      y--;
    }
    XDestroyImage(img);
    int startY = (maxY + y) / 2;

    // Find left & right edges
    Point test = (Point) { tale.width/2, startY };
    while (test.x > 0 && getPixel(test).saturation() < 0.01) {
      test.x--;
    }
    int left = test.x + 1;

    test.x = tale.width / 2;
    while (test.x < (int)tale.width && getPixel(test).saturation() < 0.01) {
      test.x++;
    }
    int right = test.x;
    if (left >= right) {
      width = 0;
      return;
    }

    // Find top & bottom (near left edge)
    img = getImage((Point){left+1,0}, 1, tale.height);
    int top = startY;
    while (top > 0 && Color::fromImage(img, 0, top).saturation() < 0.01) {
      top--;
    }
    top++;
    int bottom = startY;
    while (bottom < (int)tale.height && Color::fromImage(img, 0, bottom).saturation() < 0.01) {
      bottom++;
    }
    if (top >= bottom) {
      cerr << "Reactory height <= 0, couldn't find reactory!" << endl;
      return;
    }
    XDestroyImage(img);

    // Found area
    width = right - left;
    left += width / 128;
    right -= width / 128;
    width = right - left;
    topLeft = (Point) { left, top };
    size = (Point) { right, bottom } - topLeft;

    oldMap = getArea();
  }

  float * getArea() {
    if (width <= 0) {
      return 0;
    }
    // Get the reactory area in a grey-scale map
    // Caller is responsible for freeing the map
    XImage * img = getImage(topLeft, size.x, size.y);
    float * map = new float[size.x * size.y];
    for (int x = 0; x < size.x; x++) {
      for (int y = 0; y < size.y; y++) {
	map[x + y * width] = Color::fromImage(img, x, y).r;
      }
    }
    XDestroyImage(img);
    return map;
  }

  bool hasChanged() {
    if (width <= 0) return true;
    float * newMap = getArea();
    float diff = 0;
    for (int i = 0; i < width * size.y; i++) {
      diff += fabs(oldMap[i] - newMap[i]);
    }
    delete oldMap;
    oldMap = newMap;
    return (diff > 0);
  }

  float * filterOutlines(float * map) {
    // Filter the background from the circles
    // Caller responsible for freeing map
    float * newMap = new float[width * size.y];
    
    for (int x = 1; x < width - 1; x++) {
      for (int y = 1; y < size.y - 1; y++) {
	float center = map[x + y * width];
	float darkest = center;
	int sharp = 0;
	for (int dx = -1; dx <= 1; dx++) {
	  for (int dy = -1; dy <= 1; dy++) {
	    if (dx == 0 && dy == 0) continue;
	    float neighbour = map[x+dx + (y+dy)*width];
	    if (neighbour < darkest) {
	      darkest = neighbour;
	    }
	    float diff = center - neighbour;
	    //if (diff >= 0.05) {
	    if (diff >= 0.1) {
	      sharp++;
	    }
	  }
	}
	if (sharp > 0) {
	  // A sharp edge, try to subtract the background
	  float shifted = (center - darkest) / (1 - darkest);
	  newMap[x + y * width] = shifted;
	} else if (darkest > 0.95) {
	  // Special case for really bright blobs.  May be an edge
	  // hidden in there, but can't isolate it.  Just tone
	  // everything down.
	  newMap[x + y * width] = 0.2;
	} else {
	  // No edge, and not so bright that one could be hidden.
	  newMap[x + y * width] = 0;
	}
      }
    }

    return newMap;
  }

  float * filterBlobs(float * map, float * outlines) {
    float * blobs = copyMap(map);

    for (int x = 0; x < width; x++) {
      for (int y = 0; y < size.y; y++) {
	float outline = outlines[x + y * width];
	if (outline <= 0 || outline >= 1) continue;
	float original = map[x + y * width];
	float shifted = (original - outline) / (1 - outline);
	blobs[x + y * width] = shifted;
      }
    }

    return blobs;
  }

  void drawMap(float * map, long duration = 200000) {
    XImage * img = getImage(topLeft, size.x, size.y);
    for (int x = 0; x < width; x++) {
      for (int y = 0; y < size.y; y++) {
	float level = map[x + y * width];
	Color col = (Color) { level, level, level };
	col.toImage(img, x, y);
      }
    }

    Window w = copyWindow();
    GC gc = XCreateGC(dpy, tale.w, 0, 0);
    XPutImage(dpy, w, gc, img,  0, 0,  topLeft.x, topLeft.y,  width, size.y);
    XFlush(dpy);
    usleep(duration);
    XDestroyWindow(dpy, w);
    XFlush(dpy);

    XDestroyImage(img);
  }

  float testCircle(float * map, int cx, int cy, int steps, float threshold) {
    // Return fraction of pixels around the circumference exceeding the threshold
    float stepSize = 2 * M_PI / steps;
    float radius = width / 4.02;
    int validCount = 0;
    int lightCount = 0;
    for (int i = 0; i < steps; i++) {
      float angle = i * stepSize;
      int x = (int)(cx + radius * cosf(angle));
      if (x < 1 || x >= width) continue;
      int y = (int)(cy + radius * sinf(angle));
      if (y < 1 || y >= size.y) continue;
      validCount++;
      float lightness = map[x + y * width];
      lightness += map[x-1 + y * width];
      lightness += map[x + (y-1) * width];
      if (lightness >= threshold) {
	lightCount++;
      }
    }
    if (validCount == 0) return 0;
    return (float)lightCount / validCount;
  }

  void getCircles() {
    numCircles = 0;
    if (width == 0) return;
    float * map = getArea();
    float * outlines = filterOutlines(map);
    //float * blobs = filterBlobs(map, outlines);
    float * displayMap = copyMap(outlines);

    // Locate approximate circle centers
    struct RawCircle {
      Point center;
      float strength;
    };

    int numCenters = 0;
    RawCircle centers[1000];

    for (int x = -2; x < width + 2; x++) {
      for (int y = -2; y < size.y + 2; y++) {
	float score = testCircle(outlines, x, y, 100, 0.15);
	if (score >= 0.8 && numCenters < 1000) {
	  Point p = (Point) { x, y };
	  centers[numCenters++] = (RawCircle) { p, score };
	}
      }
    }

    // Of nearby center guesses, pick the strongest one
    for (int i = 0; i < numCenters; i++) {
      RawCircle c1 = centers[i];
      if (c1.strength == 0) continue;
      bool isStrongest = true;
      for (int j = i+1; j < numCenters && isStrongest; j++) {
	RawCircle c2 = centers[j];
	if ((c1.center - c2.center).length() > 3) continue;
	if (c2.strength > c1.strength) {
	  isStrongest = false;
	} else {
	  centers[j].strength = 0;
	}
      }
      if (isStrongest && numCircles < maxCircles) {
	circles[numCircles++] = c1.center;
      }
    }

    // Check outline strength to distinguish almost merged circles
    float radius = width / 4.02;
    int origNumCircles = numCircles;
    for (int i = 0; i < origNumCircles; i++) {
      int pixelCount = 0;
      float strength = 0;
      for (int x = 0; x < width; x++) {
	for (int y = 0; y < size.y; y++) {
	  Point p = (Point) { x, y };
	  float dist = (p - circles[i]).length();
	  if (fabsf(dist - radius) < 2) {
	    pixelCount++;
	    strength += outlines[x + y * width];
	  }
	}
      }
      strength /= pixelCount;
      //cout << "Found center at " << circles[i] << ", strength " << strength << endl;
      if (strength > 0.15 && numCircles < maxCircles) {
	// Looks like a doubled circle
	circles[numCircles++] = circles[i];
	//cout << "Doubled circle, adding extra to list" << endl;
      }
    }

    delete map;
    delete outlines;
    //delete blobs;
    delete displayMap;
  }

  void drawCircle(Point p) {
    float * map = getArea();
    int r = (int)(width / 4.02);
    for (float angle = 0; angle < 2*M_PI; angle += 0.01) {
      int x = p.x + (int)(r * cosf(angle));
      int y = p.y + (int)(r * sinf(angle));
      if (x < 0 || x >= width) continue;
      if (y < 0 || y > size.y) continue;
      map[x + y * width] = 1;
    }
    drawMap(map);
    delete map;
  }

  float * copyMap(float * map) {
    float * newMap = new float[width * size.y];
    for (int x = 0; x < width; x++) {
      for (int y = 0; y < size.y; y++) {
	newMap[x + y * width] = map[x + y * width];
      }
    }
    return newMap;
  }

  float frand() {
    return (float)random() / RAND_MAX;
  }

  CircleSet * toCircleSet(float jitter = 0) {
    CircleSet * set = new CircleSet(numCircles);
    for (int i = 0; i < numCircles; i++) {
      float x = (float)circles[i].x / width + (frand() * 2 - 1) * jitter;
      float y = (float)circles[i].y / width + (frand() * 2 - 1) * jitter;
      set->points[i] = (PointF) { x, y };
    }
    return set;
  }
  
};


int main(int argc, char *argv[]) {
  initMacro();
  srandom(time(0));
  bool reactoryFound = false;

  Reactory * reactory = new Reactory();
  while (true) {
    reactory->setBounds();
    reactory->getCircles();
    bool reactoryChanged = false;

    CircleSet * baseCircles = reactory->toCircleSet(0);
    //baseCircles->print();
    int numCircles = baseCircles->numPoints;

    if (numCircles > 0 && numCircles < 16) {
      reactoryFound = true;
      cout << numCircles << " circles found" << endl;
      baseCircles->ASCIIplot();

      Path * parent = new Path(baseCircles);
      if (argc >= 3 && strcmp(argv[1], "-steps") == 0) {
	for (int i = 2; i < argc; i++) {
	  parent->endSet->print();
	  cout << "Collisions: " << parent->endSet->numCollisions << endl;
	  cout << endl;
	  int step = atoi(argv[i]);
	  parent = new Path(parent, step);
	}
	parent->endSet->print();
	cout << "Collisions: " << parent->endSet->numCollisions << endl;
	exit(0);
      }

      int bestStep = 0;
      Path * bestPath = new Path(parent);
      float bestScore = 0;
      for (int start = 0; start < numCircles; start++) {
	if (reactory->hasChanged()) {
	  reactoryChanged = true;
	  break;
	}
	cout << start << " " << baseCircles->points[start];
	if (baseCircles->isIsolated(start)) {
	  cout << endl;
	  continue;
	}
	float avgScore = 0;
	Path * firstStep = new Path(parent, start);
	firstStep->setMaxOps(100000);
	const int numTrials = 10;
	Path * bestSubpath = new Path(firstStep);
	int bestSubpathScore = 0;
	int validTrials = 0;
	for (int i = 0; i < numTrials; i++) {
	  CircleSet * circles = reactory->toCircleSet(0.005);
	  firstStep->setStart(circles);
	  Path * best = firstStep->bestSubpath();
	  int testScore = best->score();
	  if (testScore > 0) {
	    if (testScore > bestSubpathScore) {
	      bestSubpathScore = testScore;
	      delete bestSubpath;
	      bestSubpath = new Path(best);
	    }
	    avgScore += testScore;
	    validTrials++;
	  }
	  delete best;
	  delete circles;
	}
	avgScore = avgScore * 100 / (validTrials * (numCircles - 1));
	if (avgScore > bestScore
	    || (avgScore == bestScore && bestSubpath->pathLen < bestPath->pathLen)) {
	  bestStep = start;
	  bestScore = avgScore;
	  delete bestPath;
	  bestPath = new Path(bestSubpath);
	}
	cout << " -> " << (int)avgScore << "%  eg ";
	bestSubpath->print();
	cout << " score " << bestSubpathScore << endl;
	delete bestSubpath;
      }
      cout << endl;
      /*
	cout << "Final circle positions: " << endl;
	Path * firstStep = new Path(parent, bestStep);

	PointF click = baseCircles->points[bestStep];
	for (int i = 0; i < firstStep->endSet->numPoints; i++) {
	PointF start = baseCircles->points[i];
	PointF end = firstStep->endSet->points[i];
	float startDist = start.dist(click);
	PointF motion = end - start;
	cout.setf(ios::fixed);
	cout << setprecision(3);
	cout << i << " " << start;
	if (motion.length() > 0) {
	cout << " at dist " << startDist << " moves " << motion.length() << " " << motion << " to " << end;
	}
	cout << endl;
	}

	baseCircles->ASCIIplot();
      */
  
      cout << "Best overall: ";
      bestPath->print();
      cout << " (score " << (int)bestScore << ")" << endl << endl;

      delete baseCircles;
      delete bestPath;
      delete parent;
      cout << "Waiting for reactory to change ..." << endl;
    }

    if (!reactoryChanged) {
      do {
	waitForRefresh();
	usleep(500000);
      } while (!reactory->hasChanged());
    }
  }

  endMacro();
}
