#include <iostream>
#include <iomanip>
#include <time.h>
#include <stdlib.h>
#include <math.h>

using namespace std;

struct PointF {
  float x, y;
  
  float length() {
    return sqrt(x*x + y*y);
  }

  PointF operator-(const PointF p) {
    return (PointF) { x - p.x, y - p.y };
  }

  float dist(PointF p) {
    return (*this - p).length();
  }

};

ostream& operator<<(ostream& os, PointF p)
{
  ios::fmtflags f = os.flags();
  os << setiosflags(ios::fixed) << setprecision(3);
  os << "(" << p.x << ", " << p.y << ")";
  os.flags(f);
  return os;
}

struct CircleSet {
  int numPoints;
  PointF * points;
  int numCollisions;
  static const float diameter = 0.5;
  static const float square = 1.0;

  CircleSet(int count) {
    numPoints = count;
    points = new PointF[numPoints];
    numCollisions = 0;
  }

  CircleSet(int count, int width, Point * pts) {
    numPoints = count;
    points = new PointF[numPoints];
    numCollisions = 0;
    for (int i = 0; i < numPoints; i++) {
      points[i].x = (float)pts[i].x / width;
      points[i].y = (float)pts[i].y / width;
    }
  }

  ~CircleSet() {
    delete points;
  }

  CircleSet * copy() {
    CircleSet * copy = new CircleSet(numPoints);
    for (int i = 0; i < numPoints; i++) {
      copy->points[i] = points[i];
    }
    copy->numCollisions = numCollisions;
    return copy;
  }

  bool isIsolated(int index) {
    PointF centre = points[index];
    for (int i = 0; i < numPoints; i++) {
      if (i == index)  continue;
      PointF offset = (PointF) { points[i].x - centre.x, points[i].y - centre.y };
      float dist = offset.length();
      if (dist <= diameter) {
	return false;
      }
    }
    return true;
  }

  void moveToward(int index, float tweak = 0) {
    if (numCollisions > 0)  return;
    PointF centre = points[index];
    for (int i = 0; i < numPoints; i++) {
      if (i == index)  continue;
      PointF offset = (PointF) { points[i].x - centre.x, points[i].y - centre.y };
      float dist = offset.length();
      if (dist >= diameter) {
	// Not touching
	continue;
      }
      /*
	Measurements: dist - movement
	0.139 - 0.138
	0.156 - 0.142
	0.169 - 0.153
	0.188 - 0.161
	0.191 - 0.157
	0.192 - 0.155 near
	0.192 - 0.150
	0.200 - 0.158
	0.213 - 0.163
	0.234 - 0.167
	0.234 - 0.167
	0.235 - 0.169
	0.254 - 0.168
	0.276 - 0.165
	0.278 - 0.168
	0.280 - 0.169
	0.283 - 0.162
	0.283 - 0.161
	0.307 - 0.163
	0.322 - 0.157
	0.328 - 0.151
	0.333 - 0.155
	0.344 - 0.147
	0.348 - 0.144
	0.349 - 0.141
	0.353 - 0.141
	0.355 - 0.143
	0.368 - 0.133
	0.370 - 0.131
	0.377 - 0.126
	0.390 - 0.115
	0.402 - 0.101
	0.405 - 0.102
	0.408 - 0.106
	0.411 - 0.100
	0.413 - 0.105
	0.428 - 0.082
	0.428 - 0.080
	0.429 - 0.080
	0.432 - 0.080
	0.434 - 0.083
	0.446 - 0.065
	0.453 - 0.059
	0.470 - 0.038
	0.483 - 0.023
	0.485 - 0.027
	0.489 - 0.015
       */

      float movement = dist * (diameter - dist) * 2.7 * (1 + tweak);
      float newDist = dist - movement;
      if (newDist <= 0) {
	// Collision
	numCollisions++;
	points[i] = centre;
      } else {
	// Move toward centre
	offset.x *= newDist / dist;
	offset.y *= newDist / dist;
	points[i].x = centre.x + offset.x;
	points[i].y = centre.y + offset.y;
      }
    }
  }

  void print() {
    for (int i = 0; i < numPoints; i++) {
      cout << i << ": " << points[i] << endl;
    }
    cout << endl;
  }

  void ASCIIplot() {
    int width = 20;
    int height = 10;
    char * chars = new char[width * height];

    for (int y = 0; y < height; y++) {
      for (int x = 0; x < width; x++) {
	chars[x + y * width] = ' ';
      }
    }

    for (int i = 0; i < numPoints; i++) {
      int x = points[i].x * width;
      if (x < 0) x = 0;
      if (x >= width) x = width-1;
      int y = points[i].y * height;
      if (y < 0) y = 0;
      if (y >= height) y = height-1;
      if (i <= 9) {
	chars[x + y * width] = i + '0';
      } else {
	chars[x + y * width] = (i - 10) + 'a';
      }
    }

    for (int x = 0; x < width+4; x++) {
      cout << "#";
    }
    cout << endl;
    for (int y = 0; y < height; y++) {
      cout << "# ";
      for (int x = 0; x < width; x++) {
	cout << chars[x + width * y];
      }
      cout << " #" << endl;
    }
    for (int x = 0; x < width+4; x++) {
      cout << "#";
    }
    cout << endl;
  }
};

struct Path {
  CircleSet * startSet;
  CircleSet * endSet;
  int pathLen;
  int *path;
  bool ownsMaxOps;
  int * maxOps;

  Path(CircleSet * cs) {
    startSet = cs;
    endSet = cs;
    pathLen = 0;
    ownsMaxOps = true;
    maxOps = new int;
    *maxOps = 500000;
  }

  ~Path() {
    if (pathLen > 0) {
      delete path;
      delete endSet;
    }
    if (ownsMaxOps) {
      delete maxOps;
    }
  }

  Path(Path * original, int step = -1) {
    startSet = original->startSet;
    endSet = original->endSet->copy();
    pathLen = original->pathLen;
    ownsMaxOps = false;
    maxOps = original->maxOps;
    if (step >= 0) {
      pathLen++;
    }
    if (pathLen > 0) {
      path = new int[pathLen];
    }
    for (int i = 0; i < original->pathLen; i++) {
      path[i] = original->path[i];
    }
    if (step >= 0) {
      path[pathLen-1] = step;
      if (pathLen < 20) {
	// Look for worse-case movement:
	// Circles move further than usual, unless that would make a perfect merge
	endSet->moveToward(step, 0.05);
	int collisions = endSet->numCollisions;
	if (collisions == 0 || collisions == startSet->numPoints - 1) {
	  delete endSet;
	  endSet = original->endSet->copy();
	  endSet->moveToward(step);
	}
      } else {
	endSet->numCollisions = 1;
      }
    }
  }

  void print() {
    for (int i = 0; i < pathLen; i++) {
      cout << path[i] << " ";
    }
  }

  void setStart(CircleSet * start) {
    startSet = start;
    delete endSet;
    endSet = startSet->copy();
    for (int i = 0; i < pathLen; i++) {
      endSet->moveToward(path[i]);
    }
  }

  int score() {
    return endSet->numCollisions;
  }

  float frand() {
    return (float)random() / RAND_MAX;
  }

  void setMaxOps(int max) {
    *maxOps = max;
  }

  Path * bestSubpath(int maxDepth = 1000) {
    Path * bestPath = new Path(this);
    if (score() > 0 || maxDepth <= 0 || *maxOps <= 0) {
      return bestPath;
    }
    (*maxOps)--;
    int bestScore = 0;
    int numMaxima = 1;
    bool foundPerfect = false;
    for (int i = 0; i < endSet->numPoints && !foundPerfect; i++) {
      if (endSet->isIsolated(i))  {
	continue;
      }
      Path * testPath = new Path(this, i);
      Path * bestSubpath = testPath->bestSubpath(maxDepth - 1);
      delete testPath;
      int testScore = bestSubpath->score();
      if (testScore > bestScore) {
	numMaxima = 1;
      } else if (testScore == bestScore) {
	numMaxima++;
      }
      if (testScore > bestScore) {
	delete bestPath;
	bestPath = new Path(bestSubpath);
	bestScore = testScore;
	foundPerfect = (testScore == startSet->numPoints - 1);
      }
      delete bestSubpath;
    }
    return bestPath;
  }
};

/*
int main() {
  srandom(3);

  CircleSet * circles = new CircleSet(10);
  circles->randomize();
  circles->print();

  Path * parent = new Path(circles);
  Path * best = parent->bestSubpath();
  cout << "Merged " << best->score() << " of " << (best->startSet->numPoints - 1) << " circles" << endl;
  cout << "Checked " << numOps << " paths, found sequence ";
  best->printPath();

  return 0;
}
*/
